import { CommonModule } from '@angular/common';
import { takeUntil } from 'rxjs/operators';
import { EventEmitter, Injectable, ChangeDetectorRef, Directive, Inject, Input, ViewContainerRef, NgModule } from '@angular/core';
import { BehaviorSubject, isObservable, Observable, ReplaySubject, Subject } from 'rxjs';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const defaultNgxBindIOConfig = {
    debug: false
};
/** @type {?} */
const NGX_BIND_IO_CONFIG = 'NgxBindIOConfig';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const values = new WeakMap();
/**
 * @param {?} instance
 * @return {?}
 */
function valueMap(instance) {
    /** @type {?} */
    let _valueMap = values.get(instance);
    if (!_valueMap) {
        _valueMap = new Map();
        values.set(instance, _valueMap);
    }
    return _valueMap;
}
/**
 * @param {?} target
 * @param {?} propertyKey
 * @param {?=} callback
 * @return {?}
 */
function redefineSimpleProperty(target, propertyKey, callback) {
    Object.defineProperty(target, propertyKey, {
        configurable: true,
        enumerable: true,
        /**
         * @param {?} value
         * @return {?}
         */
        set(value) {
            valueMap(this).set(propertyKey, value);
            if (callback) {
                callback(value);
            }
        },
        /**
         * @return {?}
         */
        get() {
            return valueMap(this).get(propertyKey);
        }
    });
}
/**
 * @param {?} target
 * @param {?} propertyKey
 * @param {?} descriptor
 * @param {?=} callback
 * @return {?}
 */
function redefineAccessorProperty(target, propertyKey, descriptor, callback) {
    Object.defineProperty(target, propertyKey, {
        configurable: descriptor.configurable,
        enumerable: descriptor.enumerable,
        /**
         * @param {?} value
         * @return {?}
         */
        set(value) {
            if (!descriptor.set) {
                throw new Error(`Property ${propertyKey} doesn't have a setter and cannot be written`);
            }
            descriptor.set.call(this, value);
            if (!descriptor.get) {
                callback(value);
            }
            else {
                callback(descriptor.get());
            }
        },
        /**
         * @return {?}
         */
        get() {
            if (!descriptor.get) {
                throw new Error(`Property ${propertyKey} doesn't have a getter and cannot be read`);
            }
            return descriptor.get.call(this);
        }
    });
}
/**
 * @param {?} target
 * @param {?} propertyKey
 * @return {?}
 */
function getPropDescriptor(target, propertyKey) {
    /** @type {?} */
    let originalDescriptor = Reflect.getOwnPropertyDescriptor(target, propertyKey);
    /** @type {?} */
    let setter = originalDescriptor && originalDescriptor.set;
    /** @type {?} */
    let getter = originalDescriptor && originalDescriptor.get;
    /** @type {?} */
    const value = originalDescriptor ? originalDescriptor.value : target[propertyKey];
    if (!setter || ((/** @type {?} */ (target))).__lookupSetter__(propertyKey)) {
        setter = ((/** @type {?} */ (target))).__lookupSetter__(propertyKey);
    }
    if (!getter || ((/** @type {?} */ (target))).__lookupGetter__(propertyKey)) {
        getter = ((/** @type {?} */ (target))).__lookupGetter__(propertyKey);
    }
    if (!originalDescriptor && setter && getter) {
        originalDescriptor = {
            configurable: true,
            enumerable: true,
            set: setter.bind(target),
            get: getter.bind(target)
        };
    }
    return {
        originalDescriptor: originalDescriptor,
        value,
        setter,
        getter
    };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} functionToCheck
 * @return {?}
 */
function isFunction(functionToCheck) {
    return functionToCheck && {}.toString.call(functionToCheck) === '[object Function]';
}
/**
 * @param {?} derivedCtor
 * @param {?} baseCtors
 * @return {?}
 */
function applyMixins(derivedCtor, baseCtors) {
    baseCtors.forEach(baseCtor => {
        Object.getOwnPropertyNames(baseCtor.prototype).forEach(name => {
            derivedCtor.prototype[name] = baseCtor.prototype[name];
        });
    });
    return derivedCtor;
}
/**
 * @param {?} baseCtors
 * @return {?}
 */
function Mixin(baseCtors) {
    return function (derivedCtor) {
        baseCtors.forEach(baseCtor => {
            Object.getOwnPropertyNames(baseCtor.prototype).forEach(name => {
                derivedCtor.prototype[name] = baseCtor.prototype[name];
            });
        });
    };
}
/**
 * @param {?} component
 * @param {?} rule
 * @param {?=} maxLevel
 * @return {?}
 */
function collectKeys(component, rule, maxLevel) {
    if (maxLevel !== undefined) {
        maxLevel--;
        if (maxLevel === 0) {
            return [];
        }
    }
    /** @type {?} */
    let keys = component ? Object.keys(component).filter(key => rule(component, key)) : [];
    if (component && component.__proto__) {
        keys = [...keys, ...collectKeys(component.__proto__, rule, maxLevel)];
    }
    return keys;
}
/**
 * @param {?} directive
 * @param {?} existsKeys
 * @return {?}
 */
function removeKeysUsedInAttributes(directive, existsKeys) {
    /** @type {?} */
    const viewContainerRef = (/** @type {?} */ (((/** @type {?} */ (directive)))._viewContainerRef));
    if (viewContainerRef &&
        viewContainerRef.element &&
        viewContainerRef.element.nativeElement &&
        viewContainerRef.element.nativeElement.attributes) {
        /** @type {?} */
        const attributes = (/** @type {?} */ (viewContainerRef.element.nativeElement.attributes));
        return existsKeys.filter(key => {
            /** @type {?} */
            let founded = false;
            for (let i = 0; i < attributes.length; i++) {
                /** @type {?} */
                const name = attributes
                    .item(i)
                    .name.replace(new RegExp('-', 'g'), '')
                    .toUpperCase();
                if (key.replace(new RegExp('-', 'g'), '').toUpperCase() === name) {
                    founded = true;
                }
            }
            return !founded;
        });
    }
    return existsKeys;
}
/**
 * @param {?} component
 * @return {?}
 */
function getBindIOMetadata(component) {
    if (!component['__bindIO__']) {
        ((/** @type {?} */ (component['__bindIO__']))) = {
            asHost: {
                subjects: {}
            },
            asInner: {
                subscriptions: {}
            }
        };
    }
    return (/** @type {?} */ (component['__bindIO__']));
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindInputsService {
    /**
     * BindInputs Directive
     * @param {?} directive
     * @return {?}
     */
    bindInputs(directive) {
        /** @type {?} */
        const inputs = directive.inputs;
        const { includeInputs, excludeInputs } = this.getIncludesAndExcludes(directive);
        inputs.parentKeys
            .filter(parentKey => (includeInputs.length === 0 && excludeInputs.indexOf(parentKey.toUpperCase()) === -1) ||
            (includeInputs.length !== 0 && includeInputs.indexOf(parentKey.toUpperCase()) !== -1))
            .forEach(parentKey => {
            inputs.keys
                .filter(key => (includeInputs.length === 0 && excludeInputs.indexOf(key.toUpperCase()) === -1) ||
                (includeInputs.length !== 0 && includeInputs.indexOf(key.toUpperCase()) !== -1))
                .forEach(key => {
                if (this.checkInputToBind(directive, parentKey, key)) {
                    this.bindInput(directive, parentKey, key);
                }
            });
        });
    }
    /**
     * @param {?} directive
     * @return {?}
     */
    bindObservableInputs(directive) {
        /** @type {?} */
        const inputs = directive.inputs;
        const { includeInputs, excludeInputs } = this.getIncludesAndExcludes(directive);
        inputs.parentKeys
            .filter(parentKey => (includeInputs.length === 0 && excludeInputs.indexOf(parentKey.toUpperCase()) === -1) ||
            (includeInputs.length !== 0 && includeInputs.indexOf(parentKey.toUpperCase()) !== -1))
            .forEach(parentKey => {
            inputs.keys
                .filter(key => (includeInputs.length === 0 && excludeInputs.indexOf(key.toUpperCase()) === -1) ||
                (includeInputs.length !== 0 && includeInputs.indexOf(key.toUpperCase()) !== -1))
                .forEach(key => {
                if (this.checkObservableInputToBind(directive, parentKey, key)) {
                    this.bindObservableInput(directive, parentKey, key);
                }
            });
        });
    }
    /**
     * Inputs
     * @param {?} directive
     * @param {?} parentKey
     * @param {?} key
     * @return {?}
     */
    checkKeyNameToInputBind(directive, parentKey, key) {
        return parentKey === key && parentKey[0] !== '_';
    }
    /**
     * @param {?} directive
     * @param {?} parentKey
     * @param {?} key
     * @return {?}
     */
    checkInputToBind(directive, parentKey, key) {
        /** @type {?} */
        const parentValue = getPropDescriptor(directive.parentComponent, parentKey).value || directive.parentComponent[parentKey];
        return (directive.usedInputs[parentKey] === undefined &&
            !isObservable(parentValue) &&
            this.checkKeyNameToInputBind(directive, parentKey, key));
    }
    /**
     * @param {?} directive
     * @param {?} parentKey
     * @param {?} key
     * @return {?}
     */
    bindInput(directive, parentKey, key) {
        directive.usedInputs[parentKey] = key;
        /** @type {?} */
        const descriptor = getPropDescriptor(directive.parentComponent, parentKey);
        /** @type {?} */
        const currentValue = descriptor.value || directive.parentComponent[parentKey];
        if (!getBindIOMetadata(directive.parentComponent).asHost[parentKey]) {
            getBindIOMetadata(directive.parentComponent).asHost[parentKey] = new Subject();
            delete directive.parentComponent[parentKey];
            if (descriptor !== undefined && descriptor.setter !== undefined && descriptor.getter !== undefined) {
                redefineAccessorProperty(directive.parentComponent, parentKey, descriptor.originalDescriptor, (newValue) => getBindIOMetadata(directive.parentComponent).asHost[parentKey].next(newValue));
            }
            else {
                redefineSimpleProperty(directive.parentComponent, parentKey, (newValue) => getBindIOMetadata(directive.parentComponent).asHost[parentKey].next(newValue));
            }
        }
        getBindIOMetadata(directive.parentComponent)
            .asHost[parentKey].pipe(takeUntil(directive.destroyed$))
            .subscribe(value => directive.bindValue(key, value));
        directive.parentComponent[parentKey] = currentValue;
    }
    /**
     * Observable Inputs
     * @param {?} directive
     * @param {?} parentKey
     * @param {?} key
     * @return {?}
     */
    checkKeyNameToObservableInputBind(directive, parentKey, key) {
        return parentKey === `${key}$` && parentKey[0] !== '_';
    }
    /**
     * @param {?} directive
     * @param {?} parentKey
     * @param {?} key
     * @return {?}
     */
    checkObservableInputToBind(directive, parentKey, key) {
        /** @type {?} */
        const parentValue = getPropDescriptor(directive.parentComponent, parentKey).value || directive.parentComponent[parentKey];
        return (directive.usedInputs[parentKey] === undefined &&
            isObservable(parentValue) &&
            this.checkKeyNameToObservableInputBind(directive, parentKey, key));
    }
    /**
     * @param {?} directive
     * @param {?} parentKey
     * @param {?} key
     * @return {?}
     */
    bindObservableInput(directive, parentKey, key) {
        directive.usedInputs[parentKey] = key;
        /** @type {?} */
        const descriptor = getPropDescriptor(directive.parentComponent, parentKey);
        /** @type {?} */
        const isBehaviorSubject = descriptor.value instanceof BehaviorSubject;
        /** @type {?} */
        const currentValue = descriptor.value || directive.parentComponent[parentKey];
        /** @type {?} */
        let behaviorSubjectValue;
        if (descriptor.value instanceof BehaviorSubject) {
            behaviorSubjectValue = descriptor.value.getValue();
        }
        if (!getBindIOMetadata(directive.parentComponent).asHost[parentKey]) {
            getBindIOMetadata(directive.parentComponent).asHost[parentKey] = new Subject();
            delete directive.parentComponent[parentKey];
            if (descriptor !== undefined && descriptor.setter !== undefined && descriptor.getter !== undefined) {
                redefineAccessorProperty(directive.parentComponent, parentKey, descriptor.originalDescriptor, (newValue) => getBindIOMetadata(directive.parentComponent).asHost[parentKey].next(newValue));
            }
            else {
                redefineSimpleProperty(directive.parentComponent, parentKey, (newValue) => getBindIOMetadata(directive.parentComponent).asHost[parentKey].next(newValue));
            }
        }
        getBindIOMetadata(directive.parentComponent)
            .asHost[parentKey].pipe(takeUntil(directive.destroyed$))
            .subscribe(newValue => {
            if (newValue instanceof Observable) {
                if (getBindIOMetadata(directive.component).asInner[parentKey]) {
                    getBindIOMetadata(directive.component).asInner[parentKey].unsubscribe();
                }
                getBindIOMetadata(directive.component).asInner[parentKey] = newValue
                    .pipe(takeUntil(directive.destroyed$))
                    .subscribe(value => directive.bindValue(key, value));
            }
        });
        try {
            directive.parentComponent[parentKey] = currentValue;
        }
        catch (error) { }
        if (isBehaviorSubject) {
            currentValue.next(behaviorSubjectValue);
        }
    }
    /**
     * Utils
     * @param {?} directive
     * @return {?}
     */
    getInputs(directive) {
        /** @type {?} */
        const foundedInputs = {
            parentKeys: collectKeys(directive.parentComponent, (cmp, key) => !isFunction(getPropDescriptor(directive.component, key).value) || !isFunction(directive.component[key]), 10),
            keys: directive.component
                ? [
                    ...Object.keys(directive.component).filter(key => !(getPropDescriptor(directive.component, key).value instanceof EventEmitter ||
                        directive.component[key] instanceof EventEmitter)),
                    ...collectKeys(directive.component.__proto__, (cmp, key) => !(getPropDescriptor(cmp, key).value instanceof EventEmitter), 10)
                ]
                : []
        };
        foundedInputs.parentKeys = [
            ...foundedInputs.parentKeys,
            ...foundedInputs.parentKeys
                .map(parentKey => parentKey + '$')
                .filter(parentKey => foundedInputs.parentKeys.indexOf(parentKey) === -1 &&
                (directive.parentComponent[parentKey] instanceof ReplaySubject ||
                    directive.parentComponent[parentKey] instanceof Subject ||
                    directive.parentComponent[parentKey] instanceof BehaviorSubject))
        ];
        foundedInputs.keys = removeKeysUsedInAttributes(directive, foundedInputs.keys);
        foundedInputs.parentKeys = removeKeysUsedInAttributes(directive, foundedInputs.parentKeys);
        return foundedInputs;
    }
    /**
     * @param {?} directive
     * @return {?}
     */
    getIncludesAndExcludes(directive) {
        /** @type {?} */
        const exclude = Array.isArray(directive.excludeInputs) ? directive.excludeInputs : [directive.excludeInputs];
        /** @type {?} */
        const include = Array.isArray(directive.includeInputs) ? directive.includeInputs : [directive.includeInputs];
        /** @type {?} */
        const includeIO = !directive.includeIO
            ? []
            : Array.isArray(directive.includeIO)
                ? directive.includeIO
                : [directive.includeIO];
        /** @type {?} */
        const excludeIO = !directive.excludeIO
            ? []
            : Array.isArray(directive.excludeIO)
                ? directive.excludeIO
                : [directive.excludeIO];
        /** @type {?} */
        const excludeInputs = [...exclude, ...excludeIO].map(key => key.toUpperCase());
        /** @type {?} */
        const includeInputs = [...include, ...includeIO].map(key => key.toUpperCase());
        return { includeInputs, excludeInputs };
    }
}
NgxBindInputsService.decorators = [
    { type: Injectable }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindIODebugService {
    /**
     * @param {?} _ngxBindInputsService
     */
    constructor(_ngxBindInputsService) {
        this._ngxBindInputsService = _ngxBindInputsService;
    }
    /**
     * @param {?} directive
     * @param {?} debug
     * @return {?}
     */
    showDebugInfo(directive, debug) {
        /** @type {?} */
        let notExistsOutputs = [];
        /** @type {?} */
        let notExistsInputs = [];
        if (directive.outputs &&
            directive.component &&
            directive.component.__proto__ &&
            directive.component.__proto__.constructor &&
            directive.component.__proto__.constructor.ngBaseDef &&
            directive.component.__proto__.constructor.ngBaseDef.outputs) {
            /** @type {?} */
            const ngBaseDefOutputs = Object.keys(directive.component.__proto__.constructor.ngBaseDef.outputs);
            notExistsOutputs = removeKeysUsedInAttributes(directive, ngBaseDefOutputs).filter(ngBaseDefOutput => directive.outputs.keys.indexOf(ngBaseDefOutput) === -1);
        }
        if (directive.inputs &&
            directive.component &&
            directive.component.__proto__ &&
            directive.component.__proto__.constructor &&
            directive.component.__proto__.constructor.ngBaseDef &&
            directive.component.__proto__.constructor.ngBaseDef.inputs) {
            /** @type {?} */
            const ngBaseDefInputs = Object.keys(directive.component.__proto__.constructor.ngBaseDef.inputs);
            notExistsInputs = removeKeysUsedInAttributes(directive, ngBaseDefInputs).filter(ngBaseDefInput => directive.inputs.keys.indexOf(ngBaseDefInput) === -1);
        }
        if (debug || notExistsOutputs.length > 0 || notExistsInputs.length > 0) {
            if (debug) {
                console.group('NgxBindIO: debug');
            }
            else {
                console.group('NgxBindIO: warning');
            }
            console.log('Component:', directive.component.__proto__.constructor.name, directive.component);
            if (directive.usedOutputs) {
                console.log('Outputs maping:', Object.keys(directive.usedOutputs).map(parentKey => `(${directive.usedOutputs[parentKey]})="${parentKey}($event)"`));
            }
            if (notExistsOutputs.length > 0) {
                console.log('Not initialized outputs:', notExistsOutputs);
                console.log('Not initialized outputs (text):', JSON.stringify(notExistsOutputs));
            }
            if (directive.usedInputs) {
                console.log('Inputs maping:', Object.keys(directive.usedInputs).map(parentKey => this._ngxBindInputsService.checkKeyNameToObservableInputBind(directive, parentKey, directive.usedInputs[parentKey])
                    ? `[${directive.usedInputs[parentKey]}]="${parentKey} | async"`
                    : `[${directive.usedInputs[parentKey]}]="${parentKey}"`));
            }
            if (notExistsInputs.length > 0) {
                console.log('Not initialized inputs:', notExistsInputs);
                console.log('Not initialized inputs (text):', JSON.stringify(notExistsInputs));
            }
            console.groupEnd();
        }
    }
}
NgxBindIODebugService.decorators = [
    { type: Injectable }
];
/** @nocollapse */
NgxBindIODebugService.ctorParameters = () => [
    { type: NgxBindInputsService }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindInputsDirective {
    /**
     * @param {?} _ngxBindIOConfig
     * @param {?} _viewContainerRef
     * @param {?} _ngxBindInputsService
     * @param {?} _ngxBindIODebugService
     * @param {?} _ref
     */
    constructor(_ngxBindIOConfig, _viewContainerRef, _ngxBindInputsService, _ngxBindIODebugService, _ref) {
        this._ngxBindIOConfig = _ngxBindIOConfig;
        this._viewContainerRef = _viewContainerRef;
        this._ngxBindInputsService = _ngxBindInputsService;
        this._ngxBindIODebugService = _ngxBindIODebugService;
        this._ref = _ref;
        this.excludeInputs = [];
        this.includeInputs = [];
        this.inputs = {
            keys: [],
            parentKeys: []
        };
        this.usedInputs = {};
        this.destroyed$ = new Subject();
    }
    /**
     * @return {?}
     */
    ngAfterContentInit() {
        this.component = this._viewContainerRef['_data'].componentView.component;
        this.parentComponent = ((/** @type {?} */ (this._viewContainerRef)))._view.context;
        if (this.parentComponent.$implicit !== undefined) {
            this.parentComponent = ((/** @type {?} */ (this._viewContainerRef)))._view.component;
        }
        this.inputs = this._ngxBindInputsService.getInputs(this);
        this._ngxBindInputsService.bindInputs(this);
        this._ngxBindInputsService.bindObservableInputs(this);
        /** @type {?} */
        const debug = this._ngxBindIOConfig.debug ||
            (this.bindInputs && this.bindInputs.debug) ||
            (localStorage && localStorage.getItem('debug_ngx-bind-io') === 'true');
        this._ngxBindIODebugService.showDebugInfo(this, debug);
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.destroyed$.next(true);
        this.destroyed$.complete();
    }
    /**
     * @param {?} key
     * @param {?} value
     * @return {?}
     */
    bindValue(key, value) {
        this.component[key] = value;
        this._ref.markForCheck();
    }
}
BindInputsDirective.decorators = [
    { type: Directive, args: [{
                selector: '[bindInputs]'
            },] }
];
/** @nocollapse */
BindInputsDirective.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [NGX_BIND_IO_CONFIG,] }] },
    { type: ViewContainerRef },
    { type: NgxBindInputsService },
    { type: NgxBindIODebugService },
    { type: ChangeDetectorRef }
];
BindInputsDirective.propDecorators = {
    bindInputs: [{ type: Input }],
    excludeInputs: [{ type: Input }],
    includeInputs: [{ type: Input }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindOutputsService {
    /**
     * BindOutputs Directive
     * @param {?} directive
     * @return {?}
     */
    bindOutputs(directive) {
        /** @type {?} */
        const outputs = directive.outputs;
        const { includeOutputs, excludeOutputs } = this.getIncludesAndExcludes(directive);
        outputs.parentKeys
            .filter(parentKey => (includeOutputs.length === 0 && excludeOutputs.indexOf(parentKey.toUpperCase()) === -1) ||
            (includeOutputs.length !== 0 && includeOutputs.indexOf(parentKey.toUpperCase()) !== -1))
            .forEach(parentKey => {
            outputs.keys
                .filter(key => (includeOutputs.length === 0 && excludeOutputs.indexOf(key.toUpperCase()) === -1) ||
                (includeOutputs.length !== 0 && includeOutputs.indexOf(key.toUpperCase()) !== -1))
                .forEach(key => {
                if (this.checkOutputToBind(directive, parentKey, key)) {
                    this.bindOutput(directive, parentKey, key);
                }
            });
        });
    }
    /**
     * Outputs
     * @param {?} directive
     * @param {?} parentKey
     * @param {?} key
     * @return {?}
     */
    checkKeyNameToOutputBind(directive, parentKey, key) {
        /** @type {?} */
        const outputs = directive.outputs;
        /** @type {?} */
        const keyWithFirstUpperLetter = key.length > 0 ? key.charAt(0).toUpperCase() + key.substr(1) : key;
        return ((parentKey === `on${keyWithFirstUpperLetter}` &&
            outputs.parentKeys.indexOf(`on${keyWithFirstUpperLetter}Click`) === -1) ||
            parentKey === `on${keyWithFirstUpperLetter}Click`);
    }
    /**
     * @param {?} directive
     * @param {?} parentKey
     * @param {?} key
     * @return {?}
     */
    checkOutputToBind(directive, parentKey, key) {
        /** @type {?} */
        const value = getPropDescriptor(directive.component, key).value || directive.component[key];
        return (directive.usedOutputs[parentKey] === undefined &&
            value instanceof EventEmitter &&
            this.checkKeyNameToOutputBind(directive, parentKey, key));
    }
    /**
     * @param {?} directive
     * @param {?} parentKey
     * @param {?} key
     * @return {?}
     */
    bindOutput(directive, parentKey, key) {
        directive.usedOutputs[parentKey] = key;
        directive.component[key].subscribe(value => directive.parentComponent[parentKey](value));
    }
    /**
     * Utils
     * @param {?} directive
     * @return {?}
     */
    getOutputs(directive) {
        /** @type {?} */
        const foundedOutputs = {
            parentKeys: [
                ...Object.keys(directive.parentComponent).filter(parentKey => isFunction(directive.parentComponent[parentKey])),
                ...collectKeys(directive.parentComponent.__proto__, (cmp, key) => isFunction(getPropDescriptor(cmp, key).value), 10)
            ],
            keys: directive.component
                ? [
                    ...Object.keys(directive.component).filter(key => getPropDescriptor(directive.component, key).value instanceof EventEmitter ||
                        directive.component[key] instanceof EventEmitter),
                    ...collectKeys(directive.component.__proto__, (cmp, key) => getPropDescriptor(cmp, key).value instanceof EventEmitter, 10)
                ]
                : []
        };
        foundedOutputs.keys = removeKeysUsedInAttributes(directive, foundedOutputs.keys);
        foundedOutputs.parentKeys = removeKeysUsedInAttributes(directive, foundedOutputs.parentKeys);
        return foundedOutputs;
    }
    /**
     * @param {?} directive
     * @return {?}
     */
    getIncludesAndExcludes(directive) {
        /** @type {?} */
        const exclude = Array.isArray(directive.excludeOutputs) ? directive.excludeOutputs : [directive.excludeOutputs];
        /** @type {?} */
        const include = Array.isArray(directive.includeOutputs) ? directive.includeOutputs : [directive.includeOutputs];
        /** @type {?} */
        const includeIO = !directive.includeIO
            ? []
            : Array.isArray(directive.includeIO)
                ? directive.includeIO
                : [directive.includeIO];
        /** @type {?} */
        const excludeIO = !directive.excludeIO
            ? []
            : Array.isArray(directive.excludeIO)
                ? directive.excludeIO
                : [directive.excludeIO];
        /** @type {?} */
        const excludeOutputs = [...exclude, ...excludeIO].map(key => key.toUpperCase());
        /** @type {?} */
        const includeOutputs = [...include, ...includeIO].map(key => key.toUpperCase());
        return { includeOutputs, excludeOutputs };
    }
}
NgxBindOutputsService.decorators = [
    { type: Injectable }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindIODirective {
    /**
     * @param {?} _ngxBindIOConfig
     * @param {?} _viewContainerRef
     * @param {?} _ngxBindInputsService
     * @param {?} _ngxBindOutputsService
     * @param {?} _ngxBindIODebugService
     * @param {?} _ref
     */
    constructor(_ngxBindIOConfig, _viewContainerRef, _ngxBindInputsService, _ngxBindOutputsService, _ngxBindIODebugService, _ref) {
        this._ngxBindIOConfig = _ngxBindIOConfig;
        this._viewContainerRef = _viewContainerRef;
        this._ngxBindInputsService = _ngxBindInputsService;
        this._ngxBindOutputsService = _ngxBindOutputsService;
        this._ngxBindIODebugService = _ngxBindIODebugService;
        this._ref = _ref;
        this.excludeIO = [];
        this.includeIO = [];
        this.excludeInputs = [];
        this.includeInputs = [];
        this.excludeOutputs = [];
        this.includeOutputs = [];
        this.inputs = {
            keys: [],
            parentKeys: []
        };
        this.outputs = {
            keys: [],
            parentKeys: []
        };
        this.usedInputs = {};
        this.usedOutputs = {};
        this.destroyed$ = new Subject();
    }
    /**
     * @return {?}
     */
    ngAfterContentInit() {
        this.component = this._viewContainerRef['_data'].componentView.component;
        this.parentComponent = ((/** @type {?} */ (this._viewContainerRef)))._view.context;
        if (this.parentComponent.$implicit !== undefined) {
            this.parentComponent = ((/** @type {?} */ (this._viewContainerRef)))._view.component;
        }
        this.inputs = this._ngxBindInputsService.getInputs(this);
        this.outputs = this._ngxBindOutputsService.getOutputs(this);
        this._ngxBindInputsService.bindInputs(this);
        this._ngxBindInputsService.bindObservableInputs(this);
        this._ngxBindOutputsService.bindOutputs(this);
        /** @type {?} */
        const debug = this._ngxBindIOConfig.debug ||
            (this.bindIO && this.bindIO.debug) ||
            (localStorage && localStorage.getItem('debug_ngx-bind-io') === 'true');
        this._ngxBindIODebugService.showDebugInfo(this, debug);
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.destroyed$.next(true);
        this.destroyed$.complete();
    }
    /**
     * @param {?} key
     * @param {?} value
     * @return {?}
     */
    bindValue(key, value) {
        this.component[key] = value;
        this._ref.markForCheck();
    }
}
BindIODirective.decorators = [
    { type: Directive, args: [{
                selector: '[bindIO]'
            },] }
];
/** @nocollapse */
BindIODirective.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [NGX_BIND_IO_CONFIG,] }] },
    { type: ViewContainerRef },
    { type: NgxBindInputsService },
    { type: NgxBindOutputsService },
    { type: NgxBindIODebugService },
    { type: ChangeDetectorRef }
];
BindIODirective.propDecorators = {
    bindIO: [{ type: Input }],
    excludeIO: [{ type: Input }],
    includeIO: [{ type: Input }],
    excludeInputs: [{ type: Input }],
    includeInputs: [{ type: Input }],
    excludeOutputs: [{ type: Input }],
    includeOutputs: [{ type: Input }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindOutputsDirective {
    /**
     * @param {?} _ngxBindIOConfig
     * @param {?} _viewContainerRef
     * @param {?} _ngxBindOutputsService
     * @param {?} _ngxBindIODebugService
     * @param {?} _ref
     */
    constructor(_ngxBindIOConfig, _viewContainerRef, _ngxBindOutputsService, _ngxBindIODebugService, _ref) {
        this._ngxBindIOConfig = _ngxBindIOConfig;
        this._viewContainerRef = _viewContainerRef;
        this._ngxBindOutputsService = _ngxBindOutputsService;
        this._ngxBindIODebugService = _ngxBindIODebugService;
        this._ref = _ref;
        this.excludeOutputs = [];
        this.includeOutputs = [];
        this.outputs = {
            keys: [],
            parentKeys: []
        };
        this.usedOutputs = {};
        this.destroyed$ = new Subject();
    }
    /**
     * @return {?}
     */
    ngAfterContentInit() {
        this.component = this._viewContainerRef['_data'].componentView.component;
        this.parentComponent = ((/** @type {?} */ (this._viewContainerRef)))._view.context;
        if (this.parentComponent.$implicit !== undefined) {
            this.parentComponent = ((/** @type {?} */ (this._viewContainerRef)))._view.component;
        }
        this.outputs = this._ngxBindOutputsService.getOutputs(this);
        this._ngxBindOutputsService.bindOutputs(this);
        /** @type {?} */
        const debug = this._ngxBindIOConfig.debug ||
            (this.bindOutputs && this.bindOutputs.debug) ||
            (localStorage && localStorage.getItem('debug_ngx-bind-io') === 'true');
        this._ngxBindIODebugService.showDebugInfo(this, debug);
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.destroyed$.next(true);
        this.destroyed$.complete();
    }
}
BindOutputsDirective.decorators = [
    { type: Directive, args: [{
                selector: '[bindOutputs]'
            },] }
];
/** @nocollapse */
BindOutputsDirective.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [NGX_BIND_IO_CONFIG,] }] },
    { type: ViewContainerRef },
    { type: NgxBindOutputsService },
    { type: NgxBindIODebugService },
    { type: ChangeDetectorRef }
];
BindOutputsDirective.propDecorators = {
    bindOutputs: [{ type: Input }],
    excludeOutputs: [{ type: Input }],
    includeOutputs: [{ type: Input }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindIOModule {
    /**
     * @param {?=} options
     * @return {?}
     */
    static forRoot(options) {
        return {
            ngModule: NgxBindIOModule,
            providers: [
                {
                    provide: NGX_BIND_IO_CONFIG,
                    useValue: {
                        debug: options ? options.debug : defaultNgxBindIOConfig.debug
                    }
                },
                NgxBindIODebugService,
                NgxBindInputsService,
                NgxBindOutputsService
            ]
        };
    }
}
NgxBindIOModule.decorators = [
    { type: NgModule, args: [{
                imports: [CommonModule],
                declarations: [BindInputsDirective, BindOutputsDirective, BindIODirective],
                exports: [BindInputsDirective, BindOutputsDirective, BindIODirective]
            },] }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { defaultNgxBindIOConfig, NGX_BIND_IO_CONFIG, NgxBindIOModule, BindInputsDirective, BindIODirective, BindOutputsDirective, NgxBindInputsService, NgxBindIODebugService, NgxBindOutputsService, redefineSimpleProperty, redefineAccessorProperty, getPropDescriptor, isFunction, applyMixins, Mixin, collectKeys, removeKeysUsedInAttributes, getBindIOMetadata };

//# sourceMappingURL=ngx-bind-io.js.map